import datetime as dt
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound 
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
from sqlalchemy.types import DateTime
from flask import flash, url_for, redirect, session
from wtforms.fields import StringField, HiddenField, SubmitField
from wtforms.validators import InputRequired
from openid.store import nonce
from openid.store.interface import OpenIDStore
from openid.association import Association
from openid.consumer import consumer
from openid.consumer.discover import DiscoveryFailure
from openid.extensions import sreg
from openid.yadis.constants import YADIS_HEADER_NAME, YADIS_CONTENT_TYPE

from .. import db, auth_methods, requests_session
from . import AuthMethod, AuthForm
from .models import User, Group


class OpenIDForm(AuthForm):
    url = StringField('Identifier', validators=[InputRequired()])
    submit = SubmitField('Login')


class OpenID(AuthMethod):
    name = "OpenID"

    def __init__(self, **kwargs):
        self.store = SRPOpenIDStore()

    def form(self):
        return OpenIDForm.append_field('auth_method',
                HiddenField(default=self.name))

    def login(self, form):
        con = consumer.Consumer(session, self.store)
        try:
            auth_request = con.begin(form.url.data)
        except DiscoveryFailure:
            flash("OpenID discovery error: {}".format(e), 'error')
            redirect(url_for('login'))
        else:
            # Get a nickname to identify the user with
            sreg_request = sreg.SRegRequest(required=['nickname'])
            auth_request.addExtension(sreg_request)
            # set up redirect URLs
            start_url = url_for('login')
            end_url = url_for('login',
                    auth_method=self.__class__.__name__.lower())
            auth_url = auth_request.redirectURL(start_url, end_url)
            return redirect(auth_url)

    def view(self):
        if request.method == 'POST':
            args = request.form
        elif request.method == 'GET':
            args = request.args
        if args:
            # Check OpenID response
            con = consumer.Consumer(session, self.store)
            end_url = url_for('login',
                    auth_method=self.__class__.__name__.lower())
            response = con.complete(args, end_url)
            # Check the response for the requested info
            sreg_response = {}
            if response.status == consumer.SUCCESS:
                sreg_response = sreg.SRegResponse.fromSuccessResponse(response)
                user = OpenIDUser()
                user.nickname = sreg_response['nickname']
                print(user.nickname)
                db.session.add(user)
                db.session.commit()
            else:
                flash("Error logging in with OpenID", 'error')


class SRPOpenIDStore(OpenIDStore):
    def storeAssociation(self, url, association):
        oid_user = OpenIDUser(url, association)
        db.session.add(oid_user)
        db.session.commit()

    def getAssociation(self, server_url, handle=None):
        assoc_query = OpenIDAssociation.query.filter_by(url=server_url)
        if handle is not None:
            assoc_query = assoc_query.filter_by(handle=handle)
        assoc_query = assoc_query.order_by(OpenIDAssociation.issued.desc())
        association = assoc_query.first()
        if assoc_query.expired:
            return None
        else:
            return association

    def removeAssociation(self, server_url, handle):
        query = OpenIDAssociation.query.filter_by(url=server_url, handle=handle)
        if query.exists():
            query.delete()
            db.session.commit()
            return True
        else:
            return False

    def useNonce(self, server_url, timestamp, salt):
        dt_stamp = dt.datetime.utcfromtimestamp(timestamp)
        nonce = OpenIDNonce(server_url, dt_stamp, salt)
        if nonce.valid(nonce.SKEW):
            return False
        try:
            db.session.add(nonce)
            db.session.commit()
        except IntegrityError:
            return False
        else:
            return True

    def cleanupNonces(self):
        skew = dt.timedelta(nonce.SKEW)
        return OpenIDNonce.query.filter(OpenIDNonce.valid(skew).delete())

    def cleanupAssociations(self):
        return OpenIDAssociation.query.filter(OpenIDAssociation.expired).\
                delete()


class OpenIDNonce(db.Model):
    __tablename__ = 'openidnonce'
    url = db.Column(db.String(2000), primary_key=True)
    timestamp = db.Column(DateTime(), primary_key=True)
    salt = db.Column(db.String(40), primary_key=40)

    def __init__(self, url, timestamp, salt):
        self.url = url
        self.timestamp = timestamp
        self.salt = salt

    @hybrid_method
    def valid(self, skew):
        return (self.timestamp - dt.datetime.utcnow()) <= dt.timedelta(skew)


class OpenIDAssociation(db.Model):
    __tablename__ = 'openidassociation'
    handle = db.Column(db.String(100), primary_key=True)
    url = db.Column(db.String(2000), primary_key=True)
    secret = db.Column(db.LargeBinary(128), nullable=False)
    issued = db.Column(db.DateTime(), nullable=False)
    lifetime = db.Column(db.Interval(), nullable=False)
    type_ = db.Column(db.String(50), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey('openiduser.id'))

    def __init__(self, url, association):
        self.url = url
        self.handle = association.handle
        if isinstance(association.secret):
            self.secret = association.secret.encode('utf-8')
        else:
            self.secret = association.secret
        self.issued = dt.datetime.utcfromtimestamp(association.issued)
        self.lifetime = dt.timedelta(association.lifetime)
        self.type_ = association.assoc_type

    @hybrid_property
    def expired(self):
        return (self.issued + self.interval) < dt.datetime.utcnow()


class OpenIDUser(User):
    id = db.Column(db.Integer, db.ForeignKey('user.id'), primary_key=True)
    nickname = db.Column(db.String(200), nullable=False)
    associations = db.relationship('OpenIDAssociation', collection_class=list,
            back_populates='user', order_by=OpenIDAssociation.issued.desc())

    @classmethod
    def authmethod(cls):
        return OpenID



from flask import url_for, redirect, abort, request, jsonify, Blueprint
from flask.ext.login import login_required, current_user
from sqlalchemy.orm.exc import NoResultFound

from .. import ships, systems, db
from ..models import db, Request
from ..auth import admin_permission
from ..auth.models import Division, User, Group, Pilot
from .requests import PermissionRequestListing, PersonalRequests


api = Blueprint('api', __name__)


filters = Blueprint('filters', __name__)


@api.route('/<entity_type>/')
@login_required
@admin_permission.require()
def list_entities(entity_type):
    """Return a JSON object with a list of all of the specified entity type.

    Example output::
        {
          groups: [
            {name: 'Bar', id: 1},
            {name: 'Foo', id: 0},
            {name: 'Baz', id: 20}
          ]
        }

    Order of the objects in the ``'users'`` or ``groups`` key is undefined.

    This method is only accesible to administrators.

    :param str entity_type: Either ``'user'`` or ``'group'``.
    """
    if entity_type == 'user':
        query = db.session.query(User.id, User.name)
    elif entity_type == 'group':
        query = db.session.query(Group.id, Group.name)
    else:
        abort(404)
    json_obj = {
            entity_type + 's': map(
                lambda e: {'id': e.id, 'name': e.name},
                query)
    }
    return jsonify(json_obj)


@api.route('/user/<int:user_id>/')
def user_detail(user_id):
    user = User.query.get_or_404(user_id)
    # Set up divisions
    submit = map(lambda p: p.division,
            filter(lambda p: p.permission == 'submit', user.permissions))
    review = map(lambda p: p.division,
            filter(lambda p: p.permission == 'review', user.permissions))
    pay = map(lambda p: p.division,
            filter(lambda p: p.permission == 'pay', user.permissions))
    resp = {
        'name': user.name,
        'groups': list(user.groups),
        'divisions': {
            'submit': list(set(submit)),
            'review': list(set(review)),
            'pay': list(set(pay)),
        },
        'admin': user.admin,
        'requests': user.requests,
    }
    return jsonify(resp)


@api.route('/group/<int:group_id>/')
def group_detail(group_id):
    group = Group.query.get_or_404(group_id)
    submit = map(lambda p: p.division,
            filter(lambda p: p.permission == 'submit', group.permissions))
    review = map(lambda p: p.division,
            filter(lambda p: p.permission == 'review', group.permissions))
    pay = map(lambda p: p.division,
            filter(lambda p: p.permission == 'pay', group.permissions))
    resp = {
        'name': group.name,
        'users': list(group.users),
        'divisions': {
            'submit': list(set(submit)),
            'review': list(set(review)),
            'pay': list(set(pay)),
        },
    }
    return jsonify(resp)


@api.route('/request/<int:request_id>/')
@login_required
def request_detail(request_id):
    """Get the details of a request.
    """
    request = Request.query.get_or_404(request_id)
    attrs = ('killmail_url', 'kill_timestamp', 'pilot', 'alliance',
        'corporation', 'submitter', 'division', 'status', 'base_payout',
        'payout', 'details', 'actions', 'modifiers', 'id')
    json = {}
    for attr in attrs:
        if attr == 'payout':
            json[attr] = int(request.payout)
        elif attr == 'pilot':
            json[attr] = request.pilot.name
        else:
            json[attr] = getattr(request, attr)
    json['submit_timestamp'] = request.timestamp
    return jsonify(json)


@api.route('/division/')
@login_required
@admin_permission.require()
def list_divisions():
    """List all divisions.
    """
    divisions = db.session.query(Division.id, Division.name)
    return jsonify(divisions=divisions)


@api.route('/division/<int:division_id>/')
@login_required
@admin_permission.require()
def division_detail(division_id):
    """Get the details of a division.

    :param int division_id: The ID of the division
    """
    division = Division.query.get_or_404(division_id)
    div_obj = {
            'name': division.name,
            'requests': division.requests,
    }
    permissions = {}
    for perm in ('submit', 'review', 'pay'):
        permission = division.permissions[perm]
        permissions[perm] = {
                'users': permission.individuals,
                'groups': permission.groups,
        }
    return jsonify(div_obj)


@api.route('/ships/')
@login_required
def ship_list():
    """Get an array of objects corresponding to every ship type.

    The objects have two keys, ``id`` is the integer typeID, and ``name`` is
    the name of the ship. This method is only accessible for logged in users to
    try to keep possible misuse to a minimum.
    """
    ship_objs = list(map(lambda s: {'name': s[1], 'id': s[0]},
            ships.ships.items()))
    return jsonify(ships=ship_objs)


class FiltersRequestListing(object):
    @property
    def _load_options(self):
        """Returns a sequence of
        :py:class:`~sqlalchemy.orm.strategy_options.Load` objects specifying
        which attributes to load.
        """
        return (
                db.Load(Request).load_only(
                    'id',
                    'pilot_id',
                    'corporation',
                    'alliance',
                    'ship_type',
                    'status',
                    'base_payout',
                    'kill_timestamp',
                    'timestamp',
                    'division_id',
                    'submitter_id',
                    'system',
                ),
                db.Load(Division).joinedload('name'),
                db.Load(Pilot).joinedload('name'),
                db.Load(User).joinedload('id')
        )


    def dispatch_request(self, division_id=None):
        def request_dict(request):
            payout = request.payout
            return {
                'id': request.id,
                'href': url_for('requests.request_detail',
                    request_id=request.id),
                'pilot': request.pilot.name,
                'corporation': request.corporation,
                'alliance': request.alliance,
                'ship': request.ship_type,
                'status': request.status,
                'payout': int(payout),
                'payout_str': str(payout),
                'kill_timestamp': request.kill_timestamp,
                'submit_timestamp': request.timestamp,
                'division': request.division.name,
                'submitter_id': request.submitter.id,
                'system': request.system,
                'constellation': request.constellation,
                'region': request.region,
            }

        return jsonify(requests=map(request_dict, self.requests()))


class APIRequestListing(FiltersRequestListing, PermissionRequestListing): pass


class APIPersonalRequests(FiltersRequestListing, PersonalRequests): pass


@filters.record
def register_request_lists(state):
    # Create the views
    all_requests = APIRequestListing.as_view('filter_requests_all',
            ('submit', 'review', 'pay'),
            ('evaluating', 'approved', 'paid', 'rejected', 'incomplete'))
    user_requests = APIPersonalRequests.as_view('filter_requests_own')
    pending_requests = APIRequestListing.as_view('filter_requests_pending',
            ('review',), ('evaluating', 'approved', 'incomplete'))
    pay_requests = APIRequestListing.as_view('filter_requests_pay',
            ('pay',), ('approved',))
    completed_requests = APIRequestListing.as_view('filter_requests_completed',
            ('review', 'pay'), ('paid', 'rejected'))
    # Attach the views to paths
    for prefix in state.app.request_prefixes:
        state.add_url_rule(prefix + '/', view_func=all_requests)
        state.add_url_rule(prefix + '/<int:division_id>/',
                view_func=all_requests)
        state.add_url_rule(prefix + '/personal/', view_func=user_requests)
        state.add_url_rule(prefix + '/personal/<int:division_id>/',
                view_func=user_requests)
        state.add_url_rule(prefix + '/pending/', view_func=pending_requests)
        state.add_url_rule(prefix + '/pending/<int:division_id>/',
                view_func=pending_requests)
        state.add_url_rule(prefix + '/pay/', view_func=pay_requests)
        state.add_url_rule(prefix + '/pay/<int:division_id>/',
                view_func=pay_requests)
        state.add_url_rule(prefix + '/completed/',
                view_func=completed_requests)
        state.add_url_rule(prefix + '/completed/<int:division_id>/',
                view_func=completed_requests)


@filters.route('/ship/')
@login_required
def filter_ships():
    return jsonify(ship=list(ships.ships.values()))


@filters.route('/system/')
@login_required
def filter_systems():
    return jsonify(system=list(systems.system_names.values()))


@filters.route('/constellation/')
@login_required
def filter_constellations():
    return jsonify(constellation=list(systems.constellation_names.values()))


@filters.route('/region/')
@login_required
def filter_regions():
    return jsonify(region=list(systems.region_names.values()))


def _first(o):
    return o[0]


@filters.route('/pilot/')
@login_required
def filter_pilots():
    pilots = db.session.query(Pilot.name)
    return jsonify(pilot=map(_first, pilots))


@filters.route('/corporation/')
@login_required
def filter_corps():
    corps = db.session.query(Request.corporation).distinct()
    return jsonify(corporation=map(_first, corps))


@filters.route('/alliance/')
@login_required
def filter_alliances():
    alliances = db.session.query(Request.alliance).distinct()
    return jsonify(alliance=map(_first, alliances))


@filters.route('/division/')
@login_required
def filter_divisions():
    div_names = db.session.query(Division.name)
    return jsonify(division=map(_first, div_names))



from flask import url_for, render_template, redirect, abort, flash, request,\
        Blueprint, current_app
from flask.ext.login import login_required
from flask.ext.wtf import Form
from wtforms.fields import StringField, SubmitField, HiddenField, SelectField,\
        Label
from wtforms.validators import InputRequired, AnyOf, NumberRange
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from ..models import db
from ..auth import admin_permission
from ..auth.models import Division, Permission, Entity


blueprint = Blueprint('divisions', __name__)


@blueprint.route('/')
@login_required
@admin_permission.require()
def list_divisions():
    """Show a page listing all divisions.

    Accesible only to administrators.
    """
    return render_template('divisions.html', divisions=Division.query.all())


class AddDivisionForm(Form):
    name = StringField('Division Name', validators=[InputRequired()])
    submit = SubmitField('Create Division')


@blueprint.route('/add/', methods=['GET', 'POST'])
@login_required
@admin_permission.require()
def add_division():
    """Present a form for adding a view and also process that form.

    Only accesible to adminstrators.
    """
    form = AddDivisionForm()
    if form.validate_on_submit():
        division = Division(form.name.data)
        db.session.add(division)
        db.session.commit()
        return redirect(url_for('.division_detail', division_id=division.id))
    return render_template('form.html', form=form)


class ChangeEntity(Form):
    form_id = HiddenField(default='entity')
    id_ = HiddenField()
    name = StringField()
    permission = HiddenField(validators=[AnyOf(('submit', 'review', 'pay'))])
    action = HiddenField(validators=[AnyOf(('add', 'delete'))])


class SetTransformer(Form):
    form_id = HiddenField(default='transformer')
    name = SelectField()
    kind = HiddenField()


@blueprint.route('/<int:division_id>/', methods=['GET', 'POST'])
@login_required
@admin_permission.require()
def division_detail(division_id):
    """Generate a page showing the details of a division.

    Shows which groups and individuals have been granted permissions to each
    division.

    Only accesible to administrators.

    :param int division_id: The ID number of the division
    """
    division = Division.query.get_or_404(division_id)
    ship_transformers = [('none', 'None')]
    for name in current_app.ship_urls.keys():
        ship_transformers.append((name, name))
    pilot_transformers = [('none', 'None')]
    for name in current_app.pilot_urls.keys():
        pilot_transformers.append((name, name))
    if request.method == 'POST':
        if request.form['form_id'] == 'entity':
            form = ChangeEntity()
            if form.validate():
                if form.id_.data is not None:
                    entity = Entity.query.get(form.id_.data)
                    if entity is None:
                        flash("No entity with ID #{}.".format(form.id_.data), 'error')
                else:
                    try:
                        entity = Entity.query.filter_by(name=form.name.data).one()
                    except NoResultFound:
                        flash("No entities with the name '{}' found.".
                                format(form.name.data), category='error')
                    except MultipleResultsFound:
                        flash("Multiple entities eith the name '{}' found.".
                                format(form.name.data), category='error')
                if form.action.data == 'add':
                    perm = Permission(division, form.permission.data, entity)
                    db.session.add(perm)
                    flash("Added '{}'.".format(entity))
                elif form.action.data == 'delete':
                    Permission.query.filter_by(division=division,
                            permission=form.permission.data, entity=entity).delete()
                    flash("Removed '{}' from '{}'.".format(form.permission.data,
                            entity))
        elif request.form['form_id'] == 'transformer':
            form = SetTransformer()
            if form.kind.data == 'ship':
                form.name.choices = ship_transformers
            elif form.kind.data == 'pilot':
                form.name.choices = pilot_transformers
            if form.validate():
                if form.kind.data == 'ship':
                    if form.name.data == 'none':
                        division.ship_transformer = None
                    else:
                        transformer = current_app.ship_urls[form.name.data]
                        division.ship_transformer = transformer
                elif form.kind.data == 'pilot':
                    if form.name.data == 'none':
                        division.pilot_transformer = None
                    else:
                        transformer = current_app.pilot_urls[form.name.data]
                        division.pilot_transformer = transformer
        db.session.commit()
    ship_form = SetTransformer(formdata=None)
    ship_form.name.label = Label(ship_form.name.id, 'Ship Transformers')
    ship_form.name.choices = ship_transformers
    ship_transformer = division.ship_transformer
    if ship_transformer is not None:
        ship_form.name.data = ship_transformer.name
    pilot_form = SetTransformer(formdata=None)
    pilot_form.name.label = Label(pilot_form.name.id, 'Pilot Transformers')
    pilot_form.name.choices = pilot_transformers
    pilot_transformer = division.pilot_transformer
    if pilot_transformer is not None:
        pilot_form.name.data = pilot_transformer.name
    return render_template(
            'division_detail.html',
            division=division,
            entity_form=ChangeEntity(),
            ship_form=ship_form,
            pilot_form=pilot_form
    )



from flask import View

from flask.ext.login import login_required, current_user
from .. import db
from ..models import Request, Modifier, Action
from ..auth import SubmitRequestsPermission, ReviewRequestsPermission, \
        PayoutRequestsPermission, admin_permission
from ..auth.models import Division, Pilot


class RequestListing(View):
    """Abstract class for lists of :py:class:`~evesrp.models.Request`\s.
    """

    #: The template to use for listing requests
    template = 'list_requests.html'

    decorators = [login_required]

    def requests(self, division_id=None, page=None):
        """Returns a list :py:class:`~.Request`\s belonging to
        the specified :py:class:`~.Division`, or all divisions if
        ``None``. Must be implemented by subclasses, as this is an abstract
        method.

        :param int division_id: ID number of a :py:class:`~.Division`, or
            ``None``.
        :returns: :py:class:`~.models.Request`\s
        :rtype: iterable
        """
        raise NotImplementedError()

    def dispatch_request(self, division_id=None, page=None):
        """Returns the response to requests.

        Part of the :py:class:`flask.views.View` interface.
        """
        return render_template(self.template,
                requests=self.requests(division_id, page))


class SubmittedRequestListing(RequestListing):

    def request_query(self, division_id=None):
        query = db.session.query(Request)\
                .filter(Request.submitter_id == current_user.id)
        if division_id is not None:
            query = query.filter(Request.division == division_id)
        return query


    def requests(self, division_id=None, page=None):
        if division_id is not None:
            division = Division.query.get_or_404(division_id)
            return filter(lambda r: r.division == division,
                    current_user.requests)
        else:
            return current_user.requests


class PermissionRequestListing(RequestListing):
    """Show all requests that the current user has permissions to access.

    This is used for the various permission-specific views.
    """

    def __init__(self, permissions, filter_func):
        """Create a :py:class:`PermissionRequestListing` for the given
        permissions.

        The requests can be further filtered by providing a callable via
        ``filter_func``.

        :param tuple permissions: The permissions to filter by
        :param callable filter_func: A callable taking a request as an argument
            and returning ``True`` or ``False`` if it should be included.
        """
        self.permissions = permissions
        self.filter_func = filter_func

    def request_query(self, division_id=None):
        pass

    def requests(self, division_id=None, page=None):
        if division_id is not None:
            division = Division.query.get_or_404(division_id)
            if not current_user.has_permission(self.permissions, division):
                abort(403)
            else:
                divisions = [division]
        else:
            perms = filter(lambda p: p.permission in self.permissions,
                    current_user.permissions)
            divisions = map(lambda p: p.division, perms)
        requests = OrderedDict()
        for division in divisions:
            filtered = filter(self.filter_func, division.requests)
            requests.update(map(lambda r: (r, object), filtered))
        return requests.keys()


class PermissionStatusRequests(RequestListing):
    def __init__(self, permissions, statuses):
        self.permissions = permissions
        self.statuses = statuses

    def requests(self, division_id):
        if division_id is not None:
            division = Division.query.get_or_404(division_id)
            if not current_user.has_permission(self.permissions, division):
                abort(403)
            else:
                divisions = [division]
        else:
            perms = filter(lambda p: p.permission in self.permissions,
                    current_user.permissions)
            divisions = map(lambda p: p.division, perms)
        queries = []
        for division in divisions:
            query = db.session.query(Request)\
                    .filter(Request.division_id == division.id)
            sub_queries = []
            for status in self.statuses:
                sub_query = query.filter(Request.status == status)
                sub_queries.append(sub_query)
            if len(sub_queries) > 1:
                query = sub_queries[0].union(*sub_queries[1:])
            else:
                query = sub_queries[0]
            queries.append(query)
        if len(queries) > 1:
            return queries[0].union(queries[1:])
        else:
            return queries[0]








from .killmail import CRESTMail, ZKillmail

CSRF_ENABLE = True

# This really needs to be set to something
AUTH_METHODS = []

# Default database is an in-memory SQLite DB.
SQLALCHEMY_DATABASE_URI = 'sqlite://'

# CREST killmails don't have any external dependencies
KILLMAIL_SOURCES = [ZKillmail, CRESTMail]



from ecdsa import SigningKey, VerifyingKey, NIST256p
from brave.api.client import SignedAuth, API
from sqlalchemy.orm.exc import NoResultFound
from flask import flash, url_for, redirect, abort, current_app, request
from hashlib import sha256
from binascii import unhexlify

from .. import db, requests_session
from . import AuthMethod, AuthForm
from .models import User, Group


class BraveCore(AuthMethod):
    name = 'Brave Core'

    def __init__(self, client_key, server_key, identifier,
            url='https://core.braveineve.com', **kwargs):
        self.api = API(url, identifier, client_key, server_key,
                requests_session).api
        super(BraveCore, self).__init__(self, **kwargs)

    def login(self, form):
        # Redirect to Core for the authorization token. Give URLs to return to.
        # Sidenote: At this time, Brave has nearly 0 API documentation. The
        # kinda-sorta hidden TEST Auth API documentation is more complete.
        result_url = url_for('login.auth_method_login', _external=True,
                auth_method=self.__class__.__name__.lower())
        response = self.api.core.authorize(success=result_url,
                failure=result_url)
        core_url = response['location']
        return redirect(core_url)

    def list_groups(self, user=None):
        pass

    def view(self):
        token = request.args.get('token')
        if token is not None:
            try:
                user = CoreUser.query.filter_by(token=token).one()
            except NoResultFound:
                user = CoreUser(name=None, token=token)
                db.session.add(user)
            # update user information
            info = self.api.core.info(token=token)
            user.name = info['character']['name']
            # Apply admin flag
            user.admin = user.name in self.admins
            # Sync up group membership
            for tag in info['tags']:
                try:
                    group = CoreGroup.query.filter_by(core_id=tag).one()
                except NoResultFound:
                    group = CoreGroup()
                    group.core_id = tag
                    # TODO: Figure out how to get the Group name
                    db.session.add(group)
                user.groups.append(group)
            for tag in user.groups.difference(info['tags']):
                group = CoreGroup.query.filter_by(core_id=tag).one()
                user.groups.remove(group)
            db.session.commit()
            self.login_user(user)
            # TODO Have a meaningful redirect for this
            return redirect(url_for('index'))
        else:
            flash("Login failed.")
            return redirect(url_for('login.login'))


class CoreUser(User):
    id = db.Column(db.Integer, db.ForeignKey('user.id'), primary_key=True)
    token = db.Column(db.String(100))

    @classmethod
    def authmethod(cls):
        return BraveCore


class CoreGroup(Group):
    id = db.Column(db.Integer, db.ForeignKey('group.id'), primary_key=True)
    core_id = db.Column(db.Integer, index=True)
    description = db.Column(db.Text)

    @classmethod
    def authmethod(cls):
        return BraveCore



from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.collections import attribute_mapped_collection, collection

from .. import db
from . import AuthMethod
from ..models import Action, Modifier, Request, AutoID, Timestamped


users_groups = db.Table('users_groups', db.Model.metadata,
        db.Column('user_id', db.Integer, db.ForeignKey('user.id')),
        db.Column('group_id', db.Integer, db.ForeignKey('group.id')))


class Entity(db.Model, AutoID):
    """Private class for shared functionality between :py:class:`User` and
    :py:class:`Group`.

    This class defines a number of helper methods used indirectly by User and
    Group subclasses such as automatically defining the table name and mapper
    arguments.

    You should `not` inherit fomr this class directly, and should instead
    inherit from either :py:class:`User` or :py:class:`Group`.
    """

    #: The name of the entity. Usually a nickname.
    name = db.Column(db.String(100), nullable=False)

    #: Polymorphic discriminator column.
    type_ = db.Column(db.String(50))

    #: :py:class:`Permission`\s associated specifically with this entity.
    entity_permissions = db.relationship('Permission', collection_class=set,
            back_populates='entity', lazy='dynamic')

    @declared_attr
    def __tablename__(cls):
        """SQLAlchemy late-binding attribute to set the table name.

        Implemented this way so subclasses do not need to specify a table name
        themselves.
        """
        return cls.__name__.lower()

    @declared_attr
    def __mapper_args__(cls):
        """SQLAlchemy late-binding attribute to set mapper arguments.

        Obviates subclasses from having to specify polymorphic identities.
        """
        args = {'polymorphic_identity': cls.__name__}
        if cls.__name__ == 'Entity':
            args['polymorphic_on'] = cls.type_
        return args

    @classmethod
    def authmethod(cls):
        """:rtype: class
        :returns: The :py:class:`AuthMethod` for this entity class.
        """
        return AuthMethod

    def __init__(self, name, **kwargs):
        self.name = name
        super(Entity, self).__init__(**kwargs)

    def __repr__(self):
        return "{x.__class__.__name__}('{x.name}')".format(x=self)

    def __str__(self):
        return "{x.name}".format(x=self)

    def has_permission(self, permissions, division=None):
        """Returns if this entity has been granted a permission in a division.

        If ``division`` is ``None``, this method checks if this group has the
        given permission in `any` division.

        :param permissions: The series of permissions to check
        :type permissions: iterable
        :param division: The division to check. May be ``None``.
        :type division: :py:class:`Division`
        :rtype bool:
        """
        if permissions in ('submit', 'review', 'pay'):
            permissions = (permissions,)
        perms = self.permissions.filter(Permission.permission.in_(permissions))
        return db.session.query(perms.exists()).all()[0]


class User(Entity):
    """User base class.

    Represents users who can submit, review and/or pay out requests. It also
    supplies a number of convenience methods for subclasses.
    """
    id = db.Column(db.Integer, db.ForeignKey('entity.id'), primary_key=True)

    #: If the user is an administrator. This allows the user to create and
    #: administer divisions.
    admin = db.Column(db.Boolean, nullable=False, default=False)

    #: :py:class:`~.Request`\s this user has submitted.
    requests = db.relationship(Request, back_populates='submitter')

    #: :py:class:`~.Action`\s this user has performed on requests.
    actions = db.relationship(Action, back_populates='user')

    #: :py:class:`~.Pilot`\s associated with this user.
    pilots = db.relationship('Pilot', back_populates='user',
            collection_class=set)

    #: :py:class:`Group`\s this user is a member of
    groups = db.relationship('Group', secondary=users_groups,
            back_populates='users', collection_class=set)

    notes = db.relationship('Note', back_populates='user',
            order_by='desc(Note.timestamp)', foreign_keys='Note.user_id')

    notes_made = db.relationship('Note', back_populates='noter',
            order_by='desc(Note.timestamp)', foreign_keys='Note.noter_id')

    @hybrid_property
    def permissions(self):
        """All :py:class:`Permission` objects associated with this user."""
        groups = db.session.query(users_groups.c.group_id.label('group_id'))\
                .filter(users_groups.c.user_id==self.id).subquery()
        group_perms = db.session.query(Permission)\
                .join(groups, groups.c.group_id==Permission.entity_id)
        user_perms = db.session.query(Permission)\
                .join(User)\
                .filter(User.id==self.id)
        perms = user_perms.union(group_perms)
        return perms

    @permissions.expression
    def permissions(cls):
        groups = db.select([users_groups.c.group_id])\
                .where(users_groups.c.user_id==cls.id).alias()
        group_permissions = db.select([Permission])\
                .where(Permission.entity_id.in_(groups)).alias()
        user_permissions = db.select([Permission])\
                .where(Permission.entity_id==cls.id)
        return user_permissions.union(group_permissions)

    def is_authenticated(self):
        """Part of the interface for Flask-Login."""
        return True

    def is_active(self):
        """Part of the interface for Flask-Login."""
        return True

    def is_anonymous(self):
        """Part of the interface for Flask-Login."""
        return False

    def get_id(self):
        """Part of the interface for Flask-Login."""
        return str(self.id)


class Note(db.Model, AutoID, Timestamped):
    __tablename__ = 'note'
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)
    user = db.relationship(User, back_populates='notes',
            foreign_keys=[user_id])
    content = db.Column(db.Text, nullable=False)
    noter_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)
    noter = db.relationship(User, back_populates='notes_made',
            foreign_keys=[noter_id])

    def __init__(self, user, noter, note):
        self.user = user
        self.noter = noter
        self.content = note


class Pilot(db.Model, AutoID):
    """Represents an in-game character."""
    __tablename__ = 'pilot'

    #: The name of the character
    name = db.Column(db.String(150), nullable=False)

    #: The id of the User this character belongs to.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The User this character belongs to.
    user = db.relationship(User, back_populates='pilots')

    #: The Requests filed with lossmails from this character.
    requests = db.relationship(Request, back_populates='pilot',
            collection_class=list, order_by=Request.timestamp.desc())

    def __init__(self, user, name, id_):
        """Create a new Pilot instance.

        :param user: The user this character belpongs to.
        :type user: :py:class:`~.User`
        :param str name: The name of this character.
        :param int id_: The CCP-given characterID number.
        """
        self.user = user
        self.name = name
        self.id = id_

    def __repr__(self):
        return "{x.__class__.__name__}({x.user}, '{x.name}', {x.id})".format(
                x=self)


class Group(Entity):
    """Base class for a group of users.

    Represents a group of users. Usable for granting permissions to submit,
    evaluate and pay.
    """

    id = db.Column(db.Integer, db.ForeignKey('entity.id'), primary_key=True)

    #: :py:class:`User` s that belong to this group.
    users = db.relationship(User, secondary=users_groups,
            back_populates='groups', collection_class=set)

    #: Synonym for :py:attr:`entity_permissions`
    permissions = db.synonym('entity_permissions')


class Permission(db.Model, AutoID):
    __tablename__ = 'permission'

    division_id = db.Column(db.Integer, db.ForeignKey('division.id'),
            nullable=False)

    #: The division this permission is granting access to
    division = db.relationship('Division',
            back_populates='division_permissions')

    entity_id = db.Column(db.Integer, db.ForeignKey('entity.id'),
            nullable=False)

    #: The :py:class:`Entity` being granted access
    entity = db.relationship(Entity, back_populates='entity_permissions')

    #: The permission being granted.
    permission = db.Column(db.Enum('submit', 'review', 'pay',
            name='division_permission'), nullable=False)

    def __init__(self, division, permission, entity):
        """Create a Permission object granting an entity access to a division.
        """
        self.division = division
        self.entity = entity
        self.permission = permission

    def __repr__(self):
        return ("{x.__class__.__name__}('{x.permission}', {x.entity}, "
               "{x.division})").format(x=self)


class Division(db.Model, AutoID):
    """A reimbursement division.

    A division has (possibly non-intersecting) groups of people that can submit
    requests, review requests, and pay out requests.
    """
    __tablename__ = 'division'

    #: The name of this division.
    name = db.Column(db.String(128), nullable=False)

    #: All :py:class:`Permission`\s associated with this division.
    division_permissions = db.relationship(Permission, collection_class=set,
            back_populates='division')

    #: :py:class:`Request` s filed under this division.
    requests = db.relationship(Request, back_populates='division')

    ship_transformer = db.Column(db.PickleType, nullable=True, default=None)

    pilot_transformer = db.Column(db.PickleType, nullable=True, default=None)

    @property
    def permissions(self):
        """The permissions objects for this division, mapped via their
        permission names.
        """
        class _PermProxy(object):
            def __init__(self, perms):
                self.perms = perms
            def __getitem__(self, key):
                return set(filter(lambda x: x.permission == key, self.perms))
        return _PermProxy(self.division_permissions)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "{x.__class__.__name__}('{x.name}')".format(x=self)



from collections import OrderedDict
from itertools import groupby
import re

from flask import render_template, abort, url_for, flash, Markup, request,\
    redirect, current_app, Blueprint, Markup
from flask.views import View
from flask.ext.login import login_required, current_user
from flask.ext.wtf import Form
from wtforms.fields import SelectField, SubmitField, TextAreaField, HiddenField
from wtforms.fields.html5 import URLField, DecimalField
from wtforms.validators import InputRequired, AnyOf, URL, ValidationError,\
        StopValidation

from .. import db
from ..models import Request, Modifier, Action
from ..auth import SubmitRequestsPermission, ReviewRequestsPermission, \
        PayoutRequestsPermission, admin_permission
from ..auth.models import Division, Pilot, Permission, User, Group, Note


blueprint = Blueprint('requests', __name__)


class RequestListing(View):
    """Abstract class for lists of :py:class:`~evesrp.models.Request`\s.
    """

    #: The template to use for listing requests
    template = 'list_requests.html'

    decorators = [login_required]

    def requests(self, division_id=None):
        """Returns a list :py:class:`~.Request`\s belonging to
        the specified :py:class:`~.Division`, or all divisions if
        ``None``. Must be implemented by subclasses, as this is an abstract
        method.

        :param int division_id: ID number of a :py:class:`~.Division`, or
            ``None``.
        :returns: :py:class:`~.models.Request`\s
        :rtype: iterable
        """
        raise NotImplementedError()

    def dispatch_request(self, division_id=None, page=1):
        """Returns the response to requests.

        Part of the :py:class:`flask.views.View` interface.
        """
        pager = self.requests(division_id).paginate(page, per_page=20)
        return render_template(self.template,
                pager=pager)

    @property
    def _load_options(self):
        """Returns a sequence of
        :py:class:`~sqlalchemy.orm.strategy_options.Load` objects specifying
        which attributes to load (or really any load options necessary).
        """
        return (
                db.Load(Request).load_only('id', 'pilot_id', 'division_id',
                    'system', 'ship_type', 'status', 'timestamp',
                    'base_payout'),
                db.Load(Division).joinedload('name'),
                db.Load(Pilot).joinedload('name'),
        )


class PersonalRequests(RequestListing):
    """Shows a list of all personally submitted requests and divisions they
    have permissions in.

    It will show all requests the current user has submitted.
    """

    template = 'personal.html'

    def requests(self, division_id=None):
        requests = Request.query\
                .join(User)\
                .filter(User.id==current_user.id)\
                .options(*self._load_options)
        if division_id is not None:
            requests = requests.filter(Request.division_id==division_id)
        requests = requests.order_by(Request.timestamp.desc())
        return requests


class PermissionRequestListing(RequestListing):
    """Show all requests that the current user has permissions to access.

    This is used for the various permission-specific views.
    """

    def __init__(self, permissions, statuses):
        """Create a :py:class:`PermissionRequestListing` for the given
        permissions and statuses.

        :param tuple permissions: The permissions to filter by
        :param tuple statuses: A tuple of valid statuses for requests to be in
        """
        self.permissions = permissions
        self.statuses = statuses

    def requests(self, division_id=None):
        user_perms = db.session.query(Permission.id.label('permission_id'),
                Permission.division_id.label('division_id'),
                Permission.permission.label('permission'))\
                .filter(Permission.entity==current_user)
        group_perms = db.session.query(Permission.id.label('permission_id'),
                Permission.division_id.label('division_id'),
                Permission.permission.label('permission'))\
                .join(Group)\
                .filter(Group.users.contains(current_user))
        perms = user_perms.union(group_perms)\
                .filter(Permission.permission.in_(self.permissions))
        if division_id is not None:
            perms = perms.filter(Permission.division_id==division_id)
        perms = perms.subquery()
        requests = Request.query\
                .join(perms, Request.division_id==perms.c.division_id)\
                .filter(Request.status.in_(self.statuses))\
                .order_by(Request.timestamp.desc())\
                .options(*self._load_options)
        return requests


class PayoutListing(PermissionRequestListing):

    template = 'payout.html'

    def __init__(self):
        # Just a special case of PermissionRequestListing
        super(PayoutListing, self).__init__(('pay',), ('approved',))

    def dispatch_request(self, division_id=None, page=1):
        """Returns the response to requests.

        Part of the :py:class:`flask.views.View` interface.
        """
        pager = self.requests(division_id).paginate(page, per_page=20)
        return render_template(self.template, form=ActionForm(), pager=pager)


def register_perm_request_listing(app, endpoint, path, permissions, statuses):
    """Utility function for creating :py:class:`PermissionRequestListing`
    views.

    :param app: The application to add the view to
    :type app: :py:class:`flask.Flask`
    :param str endpoint: The name of the view
    :param str path: The URL path for the view
    :param tuple permissions: Passed to
        :py:meth:`PermissionRequestListing.__init__`
    :param callable statuses: Passed to
        :py:meth:`PermissionRequestListing.__init__`
    """
    view = PermissionRequestListing.as_view(endpoint, permissions=permissions,
            statuses=statuses)
    app.add_url_rule(path, view_func=view)
    app.add_url_rule('{}<int:page>/'.format(path), view_func=view)
    app.add_url_rule('{}<int:page>/<int:division_id>'.format(path),
            view_func=view)


@blueprint.record
def register_class_views(state):
    try:
        prefixes = state.app.request_prefixes
    except AttributeError:
        prefixes = []
        state.app.request_prefixes = prefixes
    prefixes.append(state.url_prefix if state.url_prefix is not None else '')
    """Register class based views onto the requests blueprint."""
    personal_view = PersonalRequests.as_view('personal_requests')
    state.add_url_rule('/personal/', view_func=personal_view)
    state.add_url_rule('/personal/<int:page>/', view_func=personal_view)
    state.add_url_rule('/personal/<int:page>/<int:division_id>',
            view_func=personal_view)
    payout_view = PayoutListing.as_view('list_approved_requests')
    payout_url_stub = '/pay/'
    state.add_url_rule(payout_url_stub, view_func=payout_view)
    state.add_url_rule(payout_url_stub + '<int:page>/', view_func=payout_view)
    state.add_url_rule(payout_url_stub + '<int:page>/<int:division_id>/',
            view_func=payout_view)
    register_perm_request_listing(state, 'list_pending_requests',
            '/pending/', ('review',), ('evaluating', 'incomplete', 'approved'))
    register_perm_request_listing(state, 'list_completed_requests',
            '/completed/', ('review', 'pay'), ('rejected', 'paid'))


class ValidKillmail(URL):
    def __init__(self, mail_class, **kwargs):
        self.mail_class = mail_class
        super(ValidKillmail, self).__init__(**kwargs)

    def __call__(self, form, field):
        super(ValidKillmail, self).__call__(form, field)
        try:
            mail = self.mail_class(field.data)
        except ValueError as e:
            raise ValidationError(str(e)) from e
        except LookupError as e:
            raise ValidationError(str(e)) from e
        else:
            if mail.verified:
                form.killmail = mail
                raise StopValidation


def get_killmail_validators():
    """Delay accessing current_app until in a request context."""
    validators = [ValidKillmail(s) for s in current_app.killmail_sources]
    validators.append(InputRequired())
    return validators


class RequestForm(Form):
    url = URLField('Killmail URL')
    details = TextAreaField('Details', validators=[InputRequired()])
    division = SelectField('Division', coerce=int)
    submit = SubmitField('Submit')

    def validate_url(form, field):
        failures = set()
        for v in get_killmail_validators():
            try:
                v(form, field)
            except ValidationError as e:
                failures.add(str(e))
            else:
                continue
        else:
            raise ValidationError([str(e) for e in failures])


def submit_divisions(user):
    submit_perms = user.permissions\
            .filter_by(permission='submit')\
            .subquery()
    divisions = db.session.query(Division).join(submit_perms)\
            .order_by(Division.name)
    # Remove duplicates and sort divisions by name
    choices = []
    for name, group in groupby(divisions, lambda d: d.name):
        choices.append((next(group).id, name))
    return choices


@blueprint.route('/add/', methods=['GET', 'POST'])
@login_required
def submit_request():
    """Submit a :py:class:`~.models.Request`\.

    Displays a form for submitting a request and then processes the submitted
    information. Verifies that the user has the appropriate permissions to
    submit a request for the chosen division and that the killmail URL given is
    valid. Also enforces that the user submitting this requests controls the
    character from the killmail and prevents duplicate requests.
    """
    if not current_user.has_permission('submit'):
        abort(403)
    form = RequestForm()
    # Create a list of divisions this user can submit to
    form.division.choices = submit_divisions(current_user)
    if form.validate_on_submit():
        mail = form.killmail
        # Prevent submitting other people's killmails
        pilot = Pilot.query.get(mail.pilot_id)
        if not pilot or pilot not in current_user.pilots:
            flash("You can only submit killmails of characters you control",
                    'warning')
            return render_template('form.html', form=form)
        # Prevent duplicate killmails
        # The name 'request' is already used by Flask.
        # Hooray name collisions!
        srp_request = Request.query.get(mail.kill_id)
        if srp_request is None:
            division = Division.query.get(form.division.data)
            srp_request = Request(current_user, form.details.data, division,
                    mail)
            srp_request.pilot = pilot
            db.session.add(srp_request)
            db.session.commit()
            return redirect(url_for('.request_detail',
                request_id=srp_request.id))
        else:
            flash("This kill has already been submitted", 'warning')
            return redirect(url_for('.request_detail',
                request_id=srp_request.id))
    return render_template('form.html', form=form)


class ModifierForm(Form):
    id_ = HiddenField(default='modifier')
    value = DecimalField('Value')
    # TODO: add a validator for the type
    type_ = HiddenField(validators=[AnyOf(('rel-bonus', 'rel-deduct',
            'abs-bonus', 'abs-deduct'))])
    note = TextAreaField('Reason')


class VoidModifierForm(Form):
    id_ = HiddenField(default='void')
    modifier_id = HiddenField()
    void = SubmitField(Markup('x'))

    def __init__(self, modifier=None, *args, **kwargs):
        if modifier is not None:
            self.modifier_id = modifier.id
        super(VoidModifierForm, self).__init__(*args, **kwargs)


class PayoutForm(Form):
    id_ = HiddenField(default='payout')
    value = DecimalField('M ISK', validators=[InputRequired()])


class ActionForm(Form):
    id_ = HiddenField(default='action')
    note = TextAreaField('Note')
    type_ = HiddenField(default='comment', validators=[AnyOf(('rejected',
            'evaluating', 'approved', 'incomplete', 'paid', 'comment'))])


@blueprint.route('/<int:request_id>/', methods=['GET', 'POST'])
@login_required
def request_detail(request_id):
    """Renders the detail page for a :py:class:`~.models.Request`\.

    This function is currently used for `all` request detail views, including
    the reviewers and payers as well as the submitting user. It also enforces a
    the evaluation workflow, which can be seen in the diagram below. In
    addition, the payout amount can only be changed (either directly or through
    modifiers) while the request is in the 'evaluating' state.

    :param int request_id: the ID of the request

    .. digraph:: request_workflow

        rankdir="LR";

        sub [label="submitted", shape=plaintext];

        node [style="dashed, filled"];

        eval [label="evaluating", fillcolor="#fcf8e3"];
        rej [label="rejected", style="solid, filled", fillcolor="#f2dede"];
        app [label="approved", fillcolor="#d9edf7"];
        inc [label="incomplete", fillcolor="#f2dede"];
        paid [label="paid", style="solid, filled", fillcolor="#dff0d8"];

        sub -> eval;
        eval -> rej [label="R"];
        eval -> app [label="R"];
        eval -> inc [label="R"];
        rej -> eval [label="R"];
        inc -> eval [label="R, S"];
        inc -> rej [label="R"];
        app -> paid [label="P"];
        app -> eval [label="R"];
        paid -> eval [label="P"];
        paid -> app [label="P"];

    R means a reviewer can make that change, S means the submitter can make
    that change, and P means payers can make that change. Solid borders are
    terminal states.
    """
    srp_request = Request.query.get(request_id)
    review_perm = ReviewRequestsPermission(srp_request)
    pay_perm = PayoutRequestsPermission(srp_request)
    if request.method == 'POST':
        if request.form['id_'] == 'modifier':
            form = ModifierForm()
        elif request.form['id_'] == 'payout':
            form = PayoutForm()
        elif request.form['id_'] == 'action':
            form = ActionForm()
        elif request.form['id_'] == 'void':
            form = VoidModifierForm()
        else:
            abort(400)
        if form.validate():
            if srp_request.status == 'evaluating':
                if form.id_.data == 'modifier':
                    if review_perm.can():
                        mod = Modifier(srp_request, current_user, form.note.data)
                        if form.type_.data == 'rel-bonus':
                            mod.type_ = 'percentage'
                            mod.value = form.value.data
                        elif form.type_.data == 'rel-deduct':
                            mod.type_ = 'percentage'
                            mod.value = form.value.data * -1
                        elif form.type_.data == 'abs-bonus':
                            mod.type_ = 'absolute'
                            mod.value = form.value.data
                        elif form.type_.data == 'abs-deduct':
                            mod.type_ = 'absolute'
                            mod.value = form.value.data * -1
                        db.session.add(mod)
                        db.session.commit()
                    else:
                        flash("Insufficient permissions.", 'error')
                elif form.id_.data == 'payout':
                    if review_perm.can():
                        srp_request.base_payout = form.value.data
                        db.session.commit()
                    else:
                        flash("Insufficient permissions.", 'error')
                elif form.id_.data == 'void':
                    if review_perm.can():
                        modifier = Modifier.query.get(
                                int(form.modifier_id.data))
                        modifier.void(current_user)
                        db.session.commit()
                    else:
                        flash("Insufficient permissions.", 'error')
            if form.id_.data == 'action':
                # For serious, look at the diagram in the documentation before
                # tinkering around in here.
                type_ = form.type_.data
                invalid = False
                if srp_request.status == 'evaluating':
                    if type_ not in ('approved', 'rejected', 'incomplete',
                            'comment'):
                        flash("Cannot go from Evaluating to Paid", 'error')
                        invalid = True
                    elif type_ != 'comment' and not review_perm.can():
                        flash("You are not a reviewer.", 'error')
                        invalid = True
                elif srp_request.status == 'incomplete':
                    if type_ not in ('evaluating', 'rejected', 'comment'):
                        flash("Can only reject or re-evaluate.", 'error')
                        invalid = True
                    elif type_ == 'evaluating' and not (review_perm.can() or
                            srp_request.submitter == current_user):
                        flash(("You must be a reviewer or own this request to"
                               "re-evaluate."), 'error')
                        invalid = True
                    elif type_ == 'rejected' and not review_perm.can():
                        flash("You are not a reviewer.", 'error')
                        invalid = True
                elif srp_request.status == 'rejected':
                    if type_ not in ('evaluating', 'comment'):
                        flash("Can only change to Evaluating.", 'error')
                        invalid = True
                    elif type_ != 'comment' and not review_perm.can():
                        flash("You are not a reviewer.", 'error')
                        invalid = True
                elif srp_request.status == 'approved':
                    if type_ not in ('paid', 'evaluating', 'comment'):
                        flash("Can only set to Evaluating or Paid.", 'error')
                        invalid = True
                    elif type_ == 'paid' and not pay_perm.can():
                        flash("You are not a payer.", 'error')
                        invalid = True
                    elif type_ == 'evaluating' and not review_perm.can():
                        flash("You are not a reviewer.", 'error')
                        invalid = True
                elif srp_request.status == 'paid':
                    if type_ not in ('comment', 'approved', 'evaluating'):
                        flash("""Can only move to Approved or Evaluating from
                                Paid.""", 'error')
                        invalid = True
                    elif type_ != 'comment' and not pay_perm.can():
                        flash("You are not a payer.", 'error')
                        invalid = True
                if not invalid:
                    action = Action(srp_request, current_user,
                            form.note.data)
                    action.type_ = type_
                    db.session.commit()
        else:
            # TODO: Actual error handling, probably using flash()
            print(form.errors)
    # Different templates are used for different roles
    if review_perm.can():
        template = 'request_review.html'
    elif pay_perm.can():
        template = 'request_detail.html'
    elif current_user == srp_request.submitter:
        template = 'request_detail.html'
    else:
        abort(403)
    return render_template(template, srp_request=srp_request,
            modifier_form=ModifierForm(formdata=None),
            payout_form=PayoutForm(formdata=None),
            action_form=ActionForm(formdata=None),
            void_form=VoidModifierForm(formdata=None))


class DetailForm(Form):
    details = TextAreaField('Details', validators=[InputRequired()])
    submit = SubmitField('Submit')


@blueprint.route('/<int:request_id>/update/', methods=['GET', 'POST'])
@login_required
def request_detail_update(request_id):
    srp_request = Request.query.get_or_404(request_id)
    # Only the submitter can change the details, and only if it isn't finalized
    if current_user != srp_request.submitter or srp_request.finalized:
        abort(403)
    form = DetailForm()
    if form.validate_on_submit():
        archive_note = 'Old Details: ' + srp_request.details
        archive_action = Action(srp_request, current_user, archive_note)
        archive_action.type_ = 'evaluating'
        srp_request.details = form.details.data
        db.session.commit()
        return redirect(url_for('.request_detail', request_id=request_id))
    form.details.data = srp_request.details
    return render_template('form.html', form=form)


class DivisionChange(Form):
    division = SelectField('Divisions', coerce=int)
    submit = SubmitField('Submit')


@blueprint.route('/<int:request_id>/division', methods=['GET', 'POST'])
@login_required
def request_change_division(request_id):
    srp_request = Request.query.get_or_404(request_id)
    review_perm = ReviewRequestsPermission(srp_request)
    if not review_perm.can() and \
            current_user != srp_request.submitter or \
            srp_request.finalized:
        abort(403)
    form = DivisionChange()
    form.division.choices = submit_divisions(srp_request.submitter)
    if form.validate_on_submit():
        new_division = Division.query.get(form.division.data)
        archive_note = "Moving from division '{}' to division '{}'.".format(
                srp_request.division.name,
                new_division.name)
        archive_action = Action(srp_request, current_user, archive_note)
        archive_action.type_ = 'evaluating'
        srp_request.division = new_division
        db.session.commit()
        return redirect(url_for('.request_detail', request_id=request_id))
    form.division.data = srp_request.division.id
    return render_template('form.html', form=form)


@blueprint.route('/notes/<int:user_id>/', methods=['GET', 'POST'])
@login_required
def user_notes(user_id):
    if not current_user.has_permission(('review', 'pay')):
        abort(403)
    user = User.query.get_or_404(user_id)
    return render_template('notes.html', user=user)


class AddNote(Form):
    note = TextAreaField('Note',
            description=("If you have something like '#{Kill ID}', it will be "
                         "linkified to the corresponding request "
                         "(if it exists)."),
            validators=[InputRequired()])
    submit = SubmitField('Submit')


killmail_re = re.compile(r'#(\d+)')


@blueprint.route('/notes/<int:user_id>/add/', methods=['GET', 'POST'])
@login_required
def add_user_note(user_id):
    if not current_user.has_permission(('review', 'pay')):
        abort(403)
    user = User.query.get_or_404(user_id)
    form = AddNote()
    if form.validate_on_submit():
        # Linkify killmail IDs
        note_content = Markup.escape(form.note.data)
        for match in killmail_re.findall(note_content):
            kill_id = int(match)
            srp_request = db.session.query(Request.id).filter_by(id=kill_id)
            print(kill_id, srp_request.all())
            if db.session.query(srp_request.exists()):
                link = '<a href="{url}">#{kill_id}</a>'.format(
                        url=url_for('.request_detail', request_id=kill_id),
                        kill_id=kill_id)
                link = Markup(link)
                note_content = note_content.replace('#' + match, link)
        # Create the note
        note = Note(user, current_user, note_content)
        db.session.add(note)
        db.session.commit()
        return redirect(url_for('.user_notes', user_id=user_id))
    return render_template('form.html', form=form)



from flask import render_template, url_for, abort, session, redirect, request,\
        current_app, g, Blueprint
from flask.ext.login import login_required, logout_user, LoginManager
from flask.ext.principal import identity_changed, AnonymousIdentity


blueprint = Blueprint('login', __name__)


login_manager = LoginManager()


@blueprint.route('/login/', methods=['GET', 'POST'])
def login():
    """Presents the login form and processes responses from that form.

    When a POST request is recieved, this function passes control to the
    appropriate :py:meth:`login <evesrp.auth.AuthMethod.login>` method.
    """
    forms = []
    for auth_method in current_app.auth_methods:
        prefix = auth_method.__class__.__name__.lower()
        form = auth_method.form()
        forms.append((auth_method, form(prefix=prefix)))
    if request.method == 'POST':
        for auth_tuple in forms:
            if auth_tuple[1].submit.data:
                auth_method, form = auth_tuple
                break
        else:
            abort(400)
        if form.validate():
            return auth_method.login(form)
    return render_template('login.html', forms=forms)


login_manager.login_view = 'login.login'


@blueprint.route('/login/<string:auth_method>/', methods=['GET', 'POST'])
def auth_method_login(auth_method):
    """Trampoline for :py:class:`~evesrp.auth.AuthMethod`\-specific views.

    See :py:meth:`Authmethod.view <evesrp.auth.AuthMethod.view>` for more
    details.
    """
    method_map = dict(map(lambda m: (m.__class__.__name__.lower(), m),
        current_app.auth_methods))
    return method_map[auth_method].view()

auth_method_login.methods = ['GET', 'POST']


@blueprint.route('/logout/')
@login_required
def logout():
    """Logs the current user out.

    Redirects to :py:func:`.index`.
    """
    logout_user()
    for key in ('identity.name', 'identity.auth_type'):
        session.pop(key, None)
    identity_changed.send(current_app._get_current_object(),
            identity=AnonymousIdentity())
    return redirect(url_for('index'))



import re
from collections import namedtuple
from functools import partial

from flask import redirect, url_for, current_app
from flask.ext.login import current_user
import flask.ext.login as flask_login
from flask.ext.principal import Permission, UserNeed, RoleNeed,\
        identity_loaded, identity_changed, Identity, Principal
from flask.ext.wtf import Form
from wtforms.fields import SubmitField, HiddenField

from ..models import db
from ..views.login import login_manager


principal = Principal()


class AuthForm(Form):
    submit = SubmitField('Login')


class AuthMethod(object):
    name = 'Base Authentication'

    def __init__(self, admins=None, name=None, **kwargs):
        if admins is None:
            self.admins = []
        else:
            self.admins = admins
        if name is not None:
            self.name = name

    def form(self):
        """Return an instance of the form to login."""
        return AuthForm

    def login(self, form):
        """Process a validated login form.

        You must return a valid response object.
        """
        pass

    def list_groups(self, user=None):
        pass

    def view(self):
        """Optional method for providing secondary views.

        :py:func:`evesrp.views.login.auth_method_login` is configured to allow both
        GET and POST requests, and will call this method as soon as it is known
        which auth method is meant to be called. The path for this view is
        ``/login/self.__class__.__name__.lower()/``, and can be generated with
        ``url_for('login.auth_method_login',
        auth_method=self.__class__.__name__.lower())``.
        """
        return redirect(url_for('login.login'))

    @staticmethod
    def login_user(user):
        """Signal to the authentication systems that a new user has logged in.

        Handles sending the :py:data:`flask.ext.principal.identity_changed`
        signal and calling :py:func:`flask.ext.login.login_user` for you.

        :param user: The user that has been authenticated and is logging in.
        :type user: :py:class:`~models.User`
        """
        flask_login.login_user(user)
        identity_changed.send(current_app._get_current_object(),
                identity=Identity(user.id))


# Work around some circular imports
from .models import User, Group, Division


@login_manager.user_loader
def login_loader(userid):
    """Pull a user object from the database.

    This is used for loading users from existing sessions.
    """
    return User.query.get(int(userid))


# This can be confusing, so here goes. Needs really only need to be tuples, of
# some unspecified (but known) length. So, we create named tuples, and then to
# make creating them easier freeze the first argument using partial.
ReimbursementNeed = namedtuple('ReimbursementNeed', ['method', 'division'])
SubmitRequestsNeed = partial(ReimbursementNeed, 'submit')
ReviewRequestsNeed = partial(ReimbursementNeed, 'review')
PayoutRequestsNeed = partial(ReimbursementNeed, 'pay')


# Now, create Permission classes for these kinds of needs.
class SubmitRequestsPermission(Permission):
    def __init__(self, div_or_request):
        if isinstance(div_or_request, Division):
            need = SubmitRequestsNeed(div_or_request.id)
        else:
            need = SubmitRequestsNeed(div_or_request.division.id)
        super(SubmitRequestsPermission, self).__init__(need)


class ReviewRequestsPermission(Permission):
    def __init__(self, div_or_request):
        if isinstance(div_or_request, Division):
            need = ReviewRequestsNeed(div_or_request.id)
        else:
            need = ReviewRequestsNeed(div_or_request.division.id)
        super(ReviewRequestsPermission, self).__init__(need)


class PayoutRequestsPermission(Permission):
    def __init__(self, div_or_request):
        if isinstance(div_or_request, Division):
            need = PayoutRequestsNeed(div_or_request.id)
        else:
            need = PayoutRequestsNeed(div_or_request.division.id)
        super(PayoutRequestsPermission, self).__init__(need)


admin_permission = Permission(RoleNeed('admin'))


def load_user_permissions(sender, identity):
    identity.user = current_user

    if current_user.is_authenticated():
        # Set user role (see and modify their own requests)j
        identity.provides.add(UserNeed(current_user.id))

        # Set division roles
        for perm in current_user.permissions:
            identity.provides.add(ReimbursementNeed(perm.permission,
                    perm.division.id))

        # If they're an admin, set that
        if current_user.admin:
            identity.provides.add(RoleNeed('admin'))



from flask import url_for
from flask.json import JSONEncoder
from .models import Request
from .auth.models import User, Group, Division


class IterableEncoder(JSONEncoder):
    def default(self, o):
        try:
            iterator = iter(o)
        except TypeError:
            pass
        else:
            return list(o)
        return super(IterableEncoder, self).default(o)


class NamedEncoder(JSONEncoder):
    def default(self, o):
        try:
            ret = {
                'name': o.name,
                'id': o.id,
            }
        except AttributeError:
            pass
        else:
            if isinstance(o, User):
                ret['href'] = url_for('api.user_detail', user_id=o.id)
            elif isinstance(o, Group):
                ret['href'] = url_for('api.group_detail', group_id=o.id)
            elif isinstance(o, Division):
                ret['href'] = url_for('api.division_detail', division_id=o.id)
            if 'href' in ret:
                return ret
        return super(NamedEncoder, self).default(o)


class GrabbagEncoder(JSONEncoder):
    def default(self, o):
        try:
            ret = {'id': o.id}
        except AttributeError:
            pass
        else:
            done = False
            if isinstance(o, Request):
                ret['href'] = url_for('api.request_detail', request_id=o.id)
                return ret
        return super(GrabbagEncoder, self).default(o)


# Multiple inheritance FTW
class SRPEncoder(IterableEncoder, NamedEncoder, GrabbagEncoder):
    pass



import datetime as dt
from decimal import Decimal
import locale
from sqlalchemy.types import DateTime
from sqlalchemy.ext.hybrid import hybrid_property

from . import db


class AutoID(object):
    """Mixin adding a primary key integer column named 'id'."""
    id = db.Column(db.Integer, primary_key=True)


class Timestamped(object):
    """Mixin adding a timestamp column.

    The timestamp defaults to the current time.
    """
    timestamp = db.Column(DateTime, nullable=False,
            default=dt.datetime.utcnow())


action_type = db.Enum('evaluating', 'approved', 'paid', 'rejected',
        'incomplete', 'comment', name='action_type')


class Action(db.Model, AutoID, Timestamped):
    """Actions change the state of a Request.
    
    With the exception of the comment action (which does nothing), actions
    change the state of a Request.
    """

    __tablename__ = 'action'

    #: The action being taken. Must be one of: ``'evaluating'``,
    #: ``'approved'``, ``'paid'``, ``'rejected'``, ``'incomplete'``,
    #: or ``'comment'``.
    _type = db.Column(action_type, nullable=False)

    #: The ID of the :py:class:`Request` this action applies to.
    request_id = db.Column(db.Integer, db.ForeignKey('request.id'))

    #: The :py:class:`Request` this action applies to.
    request = db.relationship('Request', back_populates='actions')

    #: The ID of the :py:class:`~.User` who made this action.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The :py:class:`~.User` who made this action.
    user = db.relationship('User', back_populates='actions')

    #: Any additional notes for this action.
    note = db.Column(db.Text)

    def __init__(self, request, user, note):
        self.request = request
        self.user = user
        self.note = note
        self.timestamp = dt.datetime.utcnow()

    @property
    def type_(self):
        return self._type

    @type_.setter
    def type_(self, type_):
        if type_ != 'comment' and self.timestamp >=\
                self.request.actions[0].timestamp:
            self.request.status = type_
        self._type = type_

    def __repr__(self):
        return "{x.__class__.__name__}({x.request}, {x.user}, {x.type_})".\
                format(x=self)


class Modifier(db.Model, AutoID, Timestamped):
    """Modifiers apply bonuses or penalties to Requests.

    Modifiers come in two varieties, absolute and percentage. Absolue modifiers
    are things like "2m ISK cyno bonus" or "5m reduction for meta guns".
    Percentage modifiers apply as a percentage, such as "25% reduction for nerf
    tank" or "15% alliance logistics bonus". They can also be voided at a later
    date. The user who voided a modifier and when they did are recorded.
    """

    __tablename__ = 'modifier'

    #: What kind of modifier this is, either ``'absolute'`` or
    #: ``'percentage'``.
    type_ = db.Column(db.Enum('absolute', 'percentage',
            name='modifier_type'), nullable=False)

    #: The ID of the :py:class:`Request` this modifier applies to.
    request_id = db.Column(db.Integer, db.ForeignKey('request.id'))

    #: The :py:class:`Request` this modifier applies to.
    request = db.relationship('Request', back_populates='modifiers')

    #: The value of this modifier. If this is an absolute modifier (
    #: :py:attr:`type_` is ``'absolute'``) this is in millions of ISK. If
    #: :py:attr:`type_` is ``'percentage'``, this is the percentage of the
    #: bonus or deduction. If it's a bonus, still only set the value between
    #: 1.0 and 0.0. For example: a 20% bonus would be 0.20.
    value = db.Column(db.Float)

    #: The ID of the :py:class`~.User` who added this modifier.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The :py:class:`~.User` who added this modifier.
    user = db.relationship('User', foreign_keys=[user_id])

    #: Any notes explaining this modification.
    note = db.Column(db.Text)

    #: The ID of the :py:class:`~.User` who voided this modifier (if voided).
    voided_user_id = db.Column(db.Integer, db.ForeignKey('user.id'),
            nullable=True)

    #: The :py:class:`~.User` who voided this modifier if it has been voided.
    voided_user = db.relationship('User', foreign_keys=[voided_user_id])

    #: If this modifier has been voided, this will be the timestamp of when it
    #: was voided.
    voided_timestamp = db.Column(DateTime)

    @hybrid_property
    def voided(self):
        """Boolean of whether this modifier has been voided or not."""
        return self.voided_user is not None and \
                self.voided_timestamp is not None

    @voided.expression
    def voided(cls):
        return db.and_(
                cls.voided_user_id != None,
                cls.voided_timestamp != None
        )

    def __init__(self, request, user, note):
        self.request = request
        self.user = user
        self.note = note

    def __repr__(self):
        if self.type_ == 'absolute':
            value = "{}M ISK".format(self.value)
        else:
            value = "{}%".format(self.value)
        return ("{x.__class__.__name__}({x.request}, {x.user}, {value},"
                "{x.voided})".format(x=self, value=value))

    def void(self, user):
        """Mark this modifier as void.

        :param user: The user voiding this modifier
        :type user: :py:class:`~.User`
        """
        self.voided_user = user
        self.voided_timestamp = dt.datetime.utcnow()


class Request(db.Model, AutoID, Timestamped):
    """Requests represent SRP requests."""

    __tablename__ = 'request'

    #: The ID of the :py:class:`~.User` who submitted this request.
    submitter_id = db.Column(db.Integer, db.ForeignKey('user.id'))

    #: The :py:class:`~.User` who submitted this request.
    submitter = db.relationship('User', back_populates='requests')

    #: The ID of the :py:class`~.Division` this request was submitted to.
    division_id = db.Column(db.Integer, db.ForeignKey('division.id'),
            nullable=False)

    #: The :py:class`~.Division` this request was submitted to.
    division = db.relationship('Division', back_populates='requests')

    #: A list of :py:class:`Action`\s that have been applied to this request,
    #: sorted in the order they were applied.
    actions = db.relationship('Action', back_populates='request',
            order_by='desc(Action.timestamp)')

    #: A list of all :py:class:`Modifier`\s that have been applied to this
    #: request, regardless of wether they have been voided or not. They're
    #: sorted in the order they were added.
    modifiers = db.relationship('Modifier', back_populates='request',
            order_by='desc(Modifier.timestamp)', lazy='dynamic')

    #: The URL of the source killmail.
    killmail_url = db.Column(db.String(512), nullable=False)

    #: The ID of the :py:class:`~.Pilot` for the killmail.
    pilot_id = db.Column(db.Integer, db.ForeignKey('pilot.id'), nullable=False)

    #: The :py:class:`~.Pilot` for the killmail this request is for.
    pilot = db.relationship('Pilot', back_populates='requests')

    #: The corporation of the :py:attr:`pilot` at the time of the killmail.
    corporation = db.Column(db.String(150), nullable=False, index=True)

    #: The alliance of the :py:attr:`pilot` at the time of the killmail.
    alliance = db.Column(db.String(150), nullable=True, index=True)

    #: The type of ship that was destroyed.
    ship_type = db.Column(db.String(75), nullable=False, index=True)

    #: The date and time of when the ship was destroyed.
    kill_timestamp = db.Column(DateTime, nullable=False, index=True)

    #: The base payout for this request in millions of ISK.
    #: :py:attr:`modifiers` apply to this value.
    base_payout = db.Column(db.Float, default=0.0)

    #: Supporting information for the request.
    details = db.deferred(db.Column(db.Text))

    #: The current status of this request
    status = db.Column(action_type, nullable=False, default='evaluating')

    #: The solar system this loss occured in.
    system = db.Column(db.String(25), nullable=False, index=True)

    #: The constellation this loss occured in.
    constellation = db.Column(db.String(25), nullable=False, index=True)

    #: The region this loss occured in.
    region = db.Column(db.String(25), nullable=False, index=True)

    @property
    def payout(self):
        """The resulting payout taking all active :py:attr:`modifiers` into
        account.

        The return value is an internal class that will return different
        representations depending on the type it is being coerced to.
        :py:class:`Strings <str>` will be formatted accroding to the current
        locale with thousands separators, :py:func:`float`\s will be in
        millions of ISK, and :py:func:`ints`\s will be the total ISK value
        (equivalent to the string representation).
        """
        modifier_sum = db.session.query(db.func.sum(Modifier.value))\
                .join(Request)\
                .filter(Modifier.request_id==self.id)\
                .filter(~Modifier.voided)

        abs_mods = modifier_sum.filter(Modifier.type_=='absolute')
        per_mods = modifier_sum.filter(Modifier.type_=='percentage')
        absolute = abs_mods.one()[0]
        if absolute is None:
            absolute = 0
        percentage = per_mods.one()[0]
        if percentage is None:
            percentage = 0
        payout = self.base_payout + absolute
        payout = payout + (payout * percentage / 100)

        class _Payout(object):
            def __init__(self, payout):
                self.raw_payout = payout
                scaled = Decimal.from_float(self.raw_payout)
                scaled *= 1000000
                self.scaled_payout = scaled

            def __str__(self):
                return locale.format('%d', int(self), grouping=True)

            def __int__(self):
                return int(self.scaled_payout)

            def __float__(self):
                return self.raw_payout

        return _Payout(payout)

    @hybrid_property
    def finalized(self):
        """If this request is in a finalized status (``'paid'`` or
        ``'rejected'``).
        """
        return self.status == 'paid' or self.status == 'rejected'

    @finalized.expression
    def finalized(cls):
        return db.or_(cls.status == 'paid', cls.status == 'rejected')

    def __init__(self, submitter, details, division, killmail):
        """Create a :py:class:`Request`.

        :param submitter: The user submitting this request
        :type submitter: :py:class:`~.User`
        :param str details: Supporting details for this request
        :param division: The division this request is being submitted to
        :type division: :py:class:`~.Division`
        :param killmail: The killmail this request pertains to
        :type killmail: :py:class:`~.Killmail`
        """
        self.division = division
        self.details = details
        self.submitter = submitter
        # Pull basically everything else from the killmail object
        # The base Killmail object has an iterator defined that returns tuples
        # of Request attributes and values for those attributes
        for attr, value in killmail:
            setattr(self, attr, value)

    def __repr__(self):
        return "{x.__class__.__name__}({x.submitter}, {x.division}, {x.id})".\
                format(x=self)



import locale

import requests
from flask import Flask, current_app, g
from flask.ext.principal import identity_loaded
from flask.ext.sqlalchemy import SQLAlchemy
from flask.ext.wtf.csrf import CsrfProtect

from .sqlstats import DB_STATS


__version__ = '0.4'


requests_session = requests.Session()


# Set default locale
locale.setlocale(locale.LC_ALL, '')


db = SQLAlchemy()

# Ensure models are declared
from . import models
from .auth import models


csrf = CsrfProtect()


def create_app(**kwargs):
    app = Flask('evesrp', **kwargs)
    app.config.from_object('evesrp.default_config')

    # Register SQLAlchemy monitoring before the DB is connected
    app.before_request(sqlalchemy_before)

    db.init_app(app)

    from .views.login import login_manager
    login_manager.init_app(app)

    from .auth import principal
    principal.init_app(app)

    csrf.init_app(app)

    from .views import index, divisions, login, requests, api
    app.add_url_rule(rule='/', view_func=index)
    app.register_blueprint(divisions.blueprint, url_prefix='/divisions')
    app.register_blueprint(login.blueprint)
    app.register_blueprint(requests.blueprint)
    app.register_blueprint(api.api, url_prefix='/api')
    app.register_blueprint(api.filters, url_prefix='/api/filter')

    from .json import SRPEncoder
    app.json_encoder=SRPEncoder

    from .auth import load_user_permissions
    identity_loaded.connect(load_user_permissions, app)

    app.before_first_request(_copy_config_to_authmethods)
    app.before_first_request(_config_requests_session)
    app.before_first_request(_config_killmails)
    app.before_first_request(_copy_url_converter_config)

    return app


# SQLAlchemy performance logging
def sqlalchemy_before():
    DB_STATS.clear()
    g.DB_STATS = DB_STATS


# Auth setup
def _copy_config_to_authmethods():
    current_app.auth_methods = current_app.config['AUTH_METHODS']


# Request detail URL setup
def _copy_url_converter_config():
    current_app.ship_urls = {}
    ship_transformers = current_app.config.get('SRP_SHIP_URL_TRANSFORMERS')
    if ship_transformers is not None:
        for transformer in ship_transformers:
            current_app.ship_urls[transformer.name] = transformer
    current_app.pilot_urls = {}
    pilot_transformers = current_app.config.get('SRP_PILOT_URL_TRANSFORMERS')
    if pilot_transformers is not None:
        for transformer in pilot_transformers:
            current_app.pilot_urls[transformer.name] = transformer


# Requests session setup
def _config_requests_session():
    try:
        ua_string = current_app.config['USER_AGENT_STRING']
    except KeyError as outer_exc:
        try:
            ua_string = 'EVE-SRP/0.1 ({})'.format(
                    current_app.config['USER_AGENT_EMAIL'])
        except KeyError as inner_exc:
            raise inner_exc from outer_exc
    requests_session.headers.update({'User-Agent': ua_string})
    current_app.user_agent = ua_string


# Killmail verification
def _config_killmails():
    current_app.killmail_sources = current_app.config['KILLMAIL_SOURCES']



#!/usr/bin/env python
from setuptools import setup

setup(
    name='EVE-SRP',
    version='0.4',
    description='EVE Ship Replacement Program Helper',
    author='Will Ross',
    author_email='paxswill@paxswill.com',
    url='https://github.com/evesrp',
    packages=['evesrp', 'evesrp.auth', 'evesrp.views'],
    package_data={
        'evesrp': [
            'static/css/*.css',
            'static/css/*.css.map',
            'static/js/evesrp.min.js',
            'static/js/evesrp.min.js.map',
            'static/fonts/fontawesome-webfont.*',
            'static/ZeroClipboard.swf',
            'templates/*.html',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Flask',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Games/Entertainment'
    ],
    install_requires=[
        'Flask==0.10.1',
        'Flask-Login==0.2.10',
        'Flask-Principal==0.4.0',
        'Flask-SQLAlchemy==1.0',
        'Flask-WTF==0.9.4',
        'SQLAlchemy==0.9.3',
        'WTForms==1.0.5',
        'requests==2.2.1',
        'ecdsa==0.11',
    ],
)



class Transformer(object):

    def __init__(self, name, slug):
        self.name = name
        self.slug = slug

    def __repr__(self):
        return "{x.__class__.__name__}('{x.name}', '{x.slug}')".format(x=self)

    def __str__(self):
        return self.name

    def __call__(self, **kwargs):
        pass


class ShipTransformer(Transformer):

    def __call__(self, ship_id=None, ship_name=None, division=None):
        return self.slug.format(name=ship_name, id_=ship_id, division=division)


class PilotTransformer(Transformer):

    def __call__(self, pilot, division=None):
        return self.slug.format(name=pilot.name, id_=pilot.id,
                division=division)



from flask import redirect, url_for
from flask.ext.login import login_required


@login_required
def index():
    """The index page for EVE-SRP."""
    return redirect(url_for('requests.personal_requests'))



ships = {   582: 'Bantam',
    583: 'Condor',
    584: 'Griffin',
    585: 'Slasher',
    586: 'Probe',
    587: 'Rifter',
    588: 'Reaper',
    589: 'Executioner',
    590: 'Inquisitor',
    591: 'Tormentor',
    592: 'Navitas',
    593: 'Tristan',
    594: 'Incursus',
    595: 'Gallente Police Ship',
    596: 'Impairor',
    597: 'Punisher',
    598: 'Breacher',
    599: 'Burst',
    600: 'Minmatar Peacekeeper Ship',
    601: 'Ibis',
    602: 'Kestrel',
    603: 'Merlin',
    605: 'Heron',
    606: 'Velator',
    607: 'Imicus',
    608: 'Atron',
    609: 'Maulus',
    613: 'Devourer',
    614: 'Fury',
    615: 'Immolator',
    616: 'Medusa',
    617: 'Echo',
    618: 'Lynx',
    619: 'Swordspine',
    620: 'Osprey',
    621: 'Caracal',
    622: 'Stabber',
    623: 'Moa',
    624: 'Maller',
    625: 'Augoror',
    626: 'Vexor',
    627: 'Thorax',
    628: 'Arbitrator',
    629: 'Rupture',
    630: 'Bellicose',
    631: 'Scythe',
    632: 'Blackbird',
    633: 'Celestis',
    634: 'Exequror',
    635: 'Opux Luxury Yacht',
    638: 'Raven',
    639: 'Tempest',
    640: 'Scorpion',
    641: 'Megathron',
    642: 'Apocalypse',
    643: 'Armageddon',
    644: 'Typhoon',
    645: 'Dominix',
    648: 'Badger',
    649: 'Tayra',
    650: 'Nereus',
    651: 'Hoarder',
    652: 'Mammoth',
    653: 'Wreathe',
    654: 'Kryos',
    655: 'Epithal',
    656: 'Miasmos',
    657: 'Iteron Mark V',
    670: 'Capsule',
    671: 'Erebus',
    672: 'Caldari Shuttle',
    1233: 'Polaris Enigma Frigate',
    1896: 'Concord Police Frigate',
    1898: 'Concord SWAT Frigate',
    1900: 'Concord Army Frigate',
    1902: 'Concord Special Ops Frigate',
    1904: 'Concord Police Cruiser',
    1912: 'Concord Police Battleship',
    1914: 'Concord Special Ops Battleship',
    1916: 'Concord SWAT Battleship',
    1918: 'Concord Army Battleship',
    1944: 'Bestower',
    2006: 'Omen',
    2078: 'Zephyr',
    2161: 'Crucifier',
    2834: 'Utu',
    2836: 'Adrestia',
    2863: 'Primae',
    2998: 'Noctis',
    3514: 'Revenant',
    3516: 'Malice',
    3518: 'Vangel',
    3532: 'Echelon',
    3628: 'Nation',
    3751: 'SOCT 1',
    3753: 'SOCT 2',
    3756: 'Gnosis',
    3764: 'Leviathan',
    3766: 'Vigil',
    3768: 'Amarr Police Frigate',
    4005: 'Scorpion Ishukone Watch',
    4302: 'Oracle',
    4306: 'Naga',
    4308: 'Talos',
    4310: 'Tornado',
    4363: 'Miasmos Quafe Ultra Edition',
    4388: 'Miasmos Quafe Ultramarine Edition',
    9854: 'Polaris Inspector Frigate',
    9858: 'Polaris Centurion TEST',
    9860: 'Polaris Legatus Frigate',
    9862: 'Polaris Centurion Frigate',
    11011: 'Guardian-Vexor',
    11019: 'Cockroach',
    11129: 'Gallente Shuttle',
    11132: 'Minmatar Shuttle',
    11134: 'Amarr Shuttle',
    11172: 'Helios',
    11174: 'Keres',
    11176: 'Crow',
    11178: 'Raptor',
    11182: 'Cheetah',
    11184: 'Crusader',
    11186: 'Malediction',
    11188: 'Anathema',
    11190: 'Sentinel',
    11192: 'Buzzard',
    11194: 'Kitsune',
    11196: 'Claw',
    11198: 'Stiletto',
    11200: 'Taranis',
    11202: 'Ares',
    11365: 'Vengeance',
    11371: 'Wolf',
    11373: 'Blade',
    11375: 'Erinye',
    11377: 'Nemesis',
    11379: 'Hawk',
    11381: 'Harpy',
    11383: 'Gatherer',
    11387: 'Hyena',
    11389: 'Kishar',
    11393: 'Retribution',
    11400: 'Jaguar',
    11567: 'Avatar',
    11936: 'Apocalypse Imperial Issue',
    11938: 'Armageddon Imperial Issue',
    11940: 'Gold Magnate',
    11942: 'Silver Magnate',
    11957: 'Falcon',
    11959: 'Rook',
    11961: 'Huginn',
    11963: 'Rapier',
    11965: 'Pilgrim',
    11969: 'Arazu',
    11971: 'Lachesis',
    11978: 'Scimitar',
    11985: 'Basilisk',
    11987: 'Guardian',
    11989: 'Oneiros',
    11993: 'Cerberus',
    11995: 'Onyx',
    11999: 'Vagabond',
    12003: 'Zealot',
    12005: 'Ishtar',
    12011: 'Eagle',
    12013: 'Broadsword',
    12015: 'Muninn',
    12017: 'Devoter',
    12019: 'Sacrilege',
    12021: 'Phobos',
    12023: 'Deimos',
    12032: 'Manticore',
    12034: 'Hound',
    12036: 'Dagger',
    12038: 'Purifier',
    12042: 'Ishkur',
    12044: 'Enyo',
    12729: 'Crane',
    12731: 'Bustard',
    12733: 'Prorator',
    12735: 'Prowler',
    12743: 'Viator',
    12745: 'Occator',
    12747: 'Mastodon',
    12753: 'Impel',
    13202: 'Megathron Federate Issue',
    16227: 'Ferox',
    16229: 'Brutix',
    16231: 'Cyclone',
    16233: 'Prophecy',
    16236: 'Coercer',
    16238: 'Cormorant',
    16240: 'Catalyst',
    16242: 'Thrasher',
    17360: 'Immovable Enigma',
    17476: 'Covetor',
    17478: 'Retriever',
    17480: 'Procurer',
    17619: 'Caldari Navy Hookbill',
    17634: 'Caracal Navy Issue',
    17636: 'Raven Navy Issue',
    17703: 'Imperial Navy Slicer',
    17705: 'Khanid Navy Frigate',
    17707: 'Mordus Frigate',
    17709: 'Omen Navy Issue',
    17713: 'Stabber Fleet Issue',
    17715: 'Gila',
    17718: 'Phantasm',
    17720: 'Cynabal',
    17722: 'Vigilant',
    17726: 'Apocalypse Navy Issue',
    17728: 'Megathron Navy Issue',
    17732: 'Tempest Fleet Issue',
    17736: 'Nightmare',
    17738: 'Machariel',
    17740: 'Vindicator',
    17812: 'Republic Fleet Firetail',
    17841: 'Federation Navy Comet',
    17843: 'Vexor Navy Issue',
    17918: 'Rattlesnake',
    17920: 'Bhaalgorn',
    17922: 'Ashimmu',
    17924: 'Succubus',
    17926: 'Cruor',
    17928: 'Daredevil',
    17930: 'Worm',
    17932: 'Dramiel',
    19720: 'Revelation',
    19722: 'Naglfar',
    19724: 'Moros',
    19726: 'Phoenix',
    19744: 'Sigil',
    20125: 'Curse',
    20183: 'Providence',
    20185: 'Charon',
    20187: 'Obelisk',
    20189: 'Fenrir',
    21097: "Goru's Shuttle",
    21628: 'Guristas Shuttle',
    22428: 'Redeemer',
    22430: 'Sin',
    22436: 'Widow',
    22440: 'Panther',
    22442: 'Eos',
    22444: 'Sleipnir',
    22446: 'Vulture',
    22448: 'Absolution',
    22452: 'Heretic',
    22456: 'Sabre',
    22460: 'Eris',
    22464: 'Flycatcher',
    22466: 'Astarte',
    22468: 'Claymore',
    22470: 'Nighthawk',
    22474: 'Damnation',
    22544: 'Hulk',
    22546: 'Skiff',
    22548: 'Mackinaw',
    22852: 'Hel',
    23757: 'Archon',
    23773: 'Ragnarok',
    23911: 'Thanatos',
    23913: 'Nyx',
    23915: 'Chimera',
    23917: 'Wyvern',
    23919: 'Aeon',
    24483: 'Nidhoggur',
    24688: 'Rokh',
    24690: 'Hyperion',
    24692: 'Abaddon',
    24694: 'Maelstrom',
    24696: 'Harbinger',
    24698: 'Drake',
    24700: 'Myrmidon',
    24702: 'Hurricane',
    25560: 'Opux Dragoon Yacht',
    26840: 'Raven State Issue',
    26842: 'Tempest Tribal Issue',
    27299: 'Civilian Amarr Shuttle',
    27301: 'Civilian Caldari Shuttle',
    27303: 'Civilian Gallente Shuttle',
    27305: 'Civilian Minmatar Shuttle',
    28352: 'Rorqual',
    28606: 'Orca',
    28659: 'Paladin',
    28661: 'Kronos',
    28665: 'Vargur',
    28710: 'Golem',
    28844: 'Rhea',
    28846: 'Nomad',
    28848: 'Anshar',
    28850: 'Ark',
    29248: 'Magnate',
    29266: 'Apotheosis',
    29328: 'Amarr Media Shuttle',
    29330: 'Caldari Media Shuttle',
    29332: 'Gallente Media Shuttle',
    29334: 'Minmatar Media Shuttle',
    29336: 'Scythe Fleet Issue',
    29337: 'Augoror Navy Issue',
    29340: 'Osprey Navy Issue',
    29344: 'Exequror Navy Issue',
    29984: 'Tengu',
    29986: 'Legion',
    29988: 'Proteus',
    29990: 'Loki',
    30842: 'Interbus Shuttle',
    32207: 'Freki',
    32209: 'Mimir',
    32305: 'Armageddon Navy Issue',
    32307: 'Dominix Navy Issue',
    32309: 'Scorpion Navy Issue',
    32311: 'Typhoon Fleet Issue',
    32788: 'Cambion',
    32790: 'Etana',
    32811: 'Miasmos Amastris Edition',
    32840: 'InterBus Catalyst',
    32842: 'Intaki Syndicate Catalyst',
    32844: 'Inner Zone Shipping Catalyst',
    32846: 'Quafe Catalyst',
    32848: 'Aliastra Catalyst',
    32872: 'Algos',
    32874: 'Dragoon',
    32876: 'Corax',
    32878: 'Talwar',
    32880: 'Venture',
    32983: 'Sukuuvestaa Heron',
    32985: 'Inner Zone Shipping Imicus',
    32987: 'Sarum Magnate',
    32989: 'Vherokior Probe',
    33079: 'Hematos',
    33081: 'Taipan',
    33083: 'Violator',
    33099: 'Nefantar Thrasher',
    33151: 'Brutix Navy Issue',
    33153: 'Drake Navy Issue',
    33155: 'Harbinger Navy Issue',
    33157: 'Hurricane Fleet Issue',
    33190: 'Tash-Murkon Magnate',
    33328: "Capsule - Genolution 'Auroral' 197-variant",
    33395: 'Moracha',
    33397: 'Chremoas',
    33468: 'Astero',
    33470: 'Stratios',
    33472: 'Nestor',
    33513: 'Leopard',
    33553: 'Stratios Emergency Responder',
    33623: 'Abaddon Tash-Murkon Edition',
    33625: 'Abaddon Kador Edition',
    33627: 'Rokh Nugoeihuvi Edition',
    33629: 'Rokh Wiyrkomi Edition',
    33631: 'Maelstrom Nefantar Edition',
    33633: 'Maelstrom Krusual Edition',
    33635: 'Hyperion Aliastra Edition',
    33637: 'Hyperion Innerzone Shipping Edition',
    33639: 'Omen Kador Edition',
    33641: 'Omen Tash-Murkon Edition',
    33643: 'Caracal Nugoeihuvi Edition',
    33645: 'Caracal Wiyrkomi Edition',
    33647: 'Stabber Nefantar Edition',
    33649: 'Stabber Krusual Edition',
    33651: 'Thorax Aliastra Edition',
    33653: 'Thorax Innerzone Shipping Edition',
    33655: 'Punisher Kador Edition',
    33657: 'Punisher Tash-Murkon Edition',
    33659: 'Merlin Nugoeihuvi Edition',
    33661: 'Merlin Wiyrkomi Edition',
    33663: 'Rifter Nefantar Edition',
    33665: 'Rifter Krusual Edition',
    33667: 'Incursus Aliastra Edition',
    33669: 'Incursus Innerzone Shipping Edition',
    33673: 'Whiptail',
    33675: 'Chameleon',
    33677: 'Police Pursuit Comet',
    33683: 'Mackinaw ORE Development Edition',
    33685: 'Orca ORE Development Edition',
    33687: 'Rorqual ORE Development Edition',
    33689: 'Iteron Inner Zone Shipping Edition',
    33691: 'Tayra Wiyrkomi Edition',
    33693: 'Mammoth Nefantar Edition',
    33695: 'Bestower Tash-Murkon Edition',
    33697: 'Prospect'}
